from plone.app.contentlisting.interfaces import IContentListing
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.contentlisting.tests.base import CONTENTLISTING_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.batching.interfaces import IBatch
from plone.namedfile.file import NamedBlobImage
from Products.CMFCore.utils import getToolByName
from zope.interface.verify import verifyObject

import base64
import unittest


TEST_IMAGE_DATA = base64.b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNk+A8AAQUBAScY42YAAAAASUVORK5CYII="
)


class TestSetup(unittest.TestCase):
    layer = CONTENTLISTING_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()
        self.portal = self.layer["portal"]
        self.folder = self.portal["test-folder"]
        self.workflow = getToolByName(self.portal, "portal_workflow")
        self.catalog = getToolByName(self.portal, "portal_catalog")

    def test_able_to_add_document(self):
        # just a dummy test to see that the basics are running
        new_id = self.folder.invokeFactory("Document", "mypage")
        self.assertEqual("mypage", new_id)

    def test_simple_contentlisting(self):
        results = []
        listing = IContentListing(results)
        from plone.app.contentlisting.contentlisting import ContentListing

        self.assertTrue(isinstance(listing, ContentListing))

    def test_making_contentlisting(self):
        results = self.catalog()
        listing = IContentListing(results)
        from plone.app.contentlisting.contentlisting import ContentListing

        self.assertTrue(isinstance(listing, ContentListing))

    def test_making_contentlistingobjects(self):
        results = self.catalog()
        listing = IContentListing(results)
        from plone.app.contentlisting.catalog import CatalogContentListingObject

        self.assertTrue(isinstance(listing[0], CatalogContentListingObject))

    def test_listing_interface(self):
        self.assertTrue(
            verifyObject(
                IContentListing,
                IContentListing(self.catalog()),
            ),
        )

    def test_listing_object_interface(self):
        self.assertTrue(
            verifyObject(
                IContentListingObject,
                IContentListing(self.catalog())[0],
            ),
        )


class TestIndividualCatalogContentItems(unittest.TestCase):
    layer = CONTENTLISTING_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()
        self.portal = self.layer["portal"]
        self.folder = self.portal["test-folder"]
        self.folder.invokeFactory(
            "Document",
            "mypage",
            title="My Page",
            description="blah",
        )
        self.item = self.folder.restrictedTraverse("@@folderListing")()[0]
        self.realitem = self.folder.mypage

    def test_printing_item(self):
        self.assertEqual(
            repr(self.item),
            "<plone.app.contentlisting.catalog.CatalogContentListingObject "
            "instance at /plone/test-folder/mypage>",
        )
        self.assertEqual(
            str(self.item),
            "<plone.app.contentlisting.catalog.CatalogContentListingObject "
            "instance at /plone/test-folder/mypage>",
        )

    def test_special_getattr_with_underscore(self):
        # looking up attributes starting with _ should always raise
        # AttributeError
        self.assertRaises(AttributeError, self.item.__getattr__, "foo")

    def test_special_getattr_from_brain(self):
        # Asking for an attribute not in the contentlistingobject, should
        # defer lookup to the brain
        self.assertEqual(self.item.is_folderish, False)
        self.assertTrue(
            repr(self.item.getDataOrigin())[:35],
            "<Products.ZCatalog.Catalog.mybrains",
        )

    def test_special_getattr_from_object(self):
        # Asking for an attribute not in the contentlistingobject, should
        # defer lookup to the brain
        self.assertEqual(self.item.absolute_url(), "")
        self.assertEqual(
            repr(self.item.getDataOrigin()),
            "<Document at /plone/test-folder/mypage>",
        )

    def test_item_Title(self):
        self.assertEqual(self.item.Title(), "My Page")
        self.assertEqual(self.item.Title(), self.realitem.Title())

    def test_item_Description(self):
        self.assertEqual(self.item.Description(), "blah")
        self.assertEqual(self.item.Description(), self.realitem.Description())

    def test_item_Creator(self):
        self.assertEqual(self.item.Creator(), "test_user_1_")

    def test_item_getURL(self):
        self.assertEqual(
            self.item.getURL(),
            "http://nohost/plone/test-folder/mypage",
        )
        self.assertEqual(self.item.getURL(), self.realitem.absolute_url())

    def test_item_getSize(self):
        self.assertEqual(self.item.getSize().upper(), "0 KB")

    def test_item_reviewState(self):
        wftool = getToolByName(self.realitem, "portal_workflow")
        wf = wftool.getInfoFor(self.realitem, "review_state")
        self.assertEqual(self.item.review_state(), wf)

    def test_item_Type(self):
        self.assertEqual(self.item.Type(), "Page")
        self.assertEqual(self.item.Type().domain, "plone")

    def test_appendViewAction(self):
        # checking that we append the view action to urls when needed
        self.assertEqual(self.item.appendViewAction(), "")
        self.folder.invokeFactory(
            "Image",
            "myimage",
            title="My Image",
            description="blah",
        )
        self.item = self.folder.restrictedTraverse("@@contentlisting")()[1]
        self.assertEqual(self.item.appendViewAction(), "/view")

    def test_item_ContentTypeClass(self):
        # checking the that we print nice strings for css class identifiers
        self.assertEqual(self.item.ContentTypeClass(), "contenttype-document")

    def test_comparision(self):
        self.assertEqual(IContentListingObject(self.folder.mypage), self.item)

    def test_containment(self):
        # we can test containment for normal content objects against
        # contentlistings
        self.assertTrue(
            self.folder.mypage in self.folder.restrictedTraverse("@@contentlisting")(),
        )


class TestIndividualRealContentItems(unittest.TestCase):
    layer = CONTENTLISTING_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()
        self.portal = self.layer["portal"]
        self.folder = self.portal["test-folder"]
        self.folder.invokeFactory(
            "Document",
            "mypage",
            title="My Page",
            description="blah",
        )
        self.item = IContentListingObject(self.folder.mypage)
        self.realitem = self.folder.mypage

    def test_printing_item(self):
        self.assertEqual(
            repr(self.item),
            "<plone.app.contentlisting.realobject.RealContentListingObject "
            "instance at /plone/test-folder/mypage>",
        )
        self.assertEqual(
            str(self.item),
            "<plone.app.contentlisting.realobject.RealContentListingObject "
            "instance at /plone/test-folder/mypage>",
        )

    def test_special_getattr_with_underscore(self):
        # looking up attributes starting with _ should always raise
        # AttributeError
        self.assertRaises(AttributeError, self.item.__getattr__, "foo")

    def test_item_getattr_acquisition(self):
        # absolute_url needs acquisition to work
        # on the object
        self.assertEqual(
            self.item._realobject.absolute_url(),
            "http://nohost/plone/test-folder/mypage",
        )
        # on the contentlisting object
        self.assertEqual(
            self.item.absolute_url(), "http://nohost/plone/test-folder/mypage"
        )

    def test_item_Title(self):
        self.assertEqual(self.item.Title(), "My Page")
        self.assertEqual(self.item.Title(), self.realitem.Title())

    def test_item_Description(self):
        self.assertEqual(self.item.Description(), "blah")
        self.assertEqual(self.item.Description(), self.realitem.Description())

    def test_item_Creator(self):
        self.assertEqual(self.item.Creator(), "test_user_1_")

    def test_item_getURL(self):
        self.assertEqual(
            self.item.getURL(),
            "http://nohost/plone/test-folder/mypage",
        )
        self.assertEqual(self.item.getURL(), self.realitem.absolute_url())

    def test_item_getSize(self):
        self.assertEqual(self.item.getSize().upper(), "0 KB")

    def test_item_reviewState(self):
        wftool = getToolByName(self.realitem, "portal_workflow")
        wf = wftool.getInfoFor(self.realitem, "review_state")
        self.assertEqual(self.item.review_state(), wf)

    def test_item_Type(self):
        self.assertEqual(self.item.Type(), "Page")
        self.assertEqual(self.item.Type().domain, "plone")

    def test_item_ContentTypeClass(self):
        # checking the that we print nice strings for css class identifiers
        self.assertEqual(self.item.ContentTypeClass(), "contenttype-document")

    def test_item_uuid(self):
        uuid = self.item.uuid()
        assert uuid
        assert uuid != self.item.getPath()

    def test_item_get_none(self):
        self.realitem.test_none = None
        try:
            self.assertEqual(self.item.test_none, None)
        except AttributeError:
            self.fail(
                "Accessing attributes which return ``None`` should not "
                "result in an AttributeError."
            )

    def test_item_image_scales(self):
        self.folder.invokeFactory(
            "Image",
            "myimage",
            title="My Image",
        )
        myimage = self.folder.myimage
        myimage.image = NamedBlobImage(
            data=TEST_IMAGE_DATA,
            filename="test.png",
        )
        item = IContentListingObject(myimage)
        image_scales = item.image_scales()
        self.assertIn("download", image_scales["image"][0])


class TestFolderContents(unittest.TestCase):
    """Testing that the folder contents browserview works and behaves
    as it should.
    """

    layer = CONTENTLISTING_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()
        self.portal = self.layer["portal"]
        self.folder = self.portal["test-folder"]

    def test_empty_folder_contents(self):
        contentlisting = self.folder.restrictedTraverse("@@contentlisting")()
        self.assertEqual(len(contentlisting), 0)
        self.assertEqual(contentlisting.actual_result_count, 0)

    def test_item_in_folder_contents(self):
        # adding a new page, adds to the length of folder contents
        self.folder.invokeFactory("Document", "mypage")
        contentlisting = self.folder.restrictedTraverse("@@contentlisting")()
        self.assertEqual(len(contentlisting), 1)
        self.assertEqual(contentlisting.actual_result_count, 1)

    def test_folder_contents(self):
        # call the generic folder contents browserview. Check that it makes
        # the results a contentlisting, regardless of batching
        self.folder.invokeFactory("Document", "mypage")
        contentlisting = self.folder.restrictedTraverse("@@contentlisting")()
        self.assertTrue(verifyObject(IContentListing, contentlisting))

    def test_batching_folder_contents(self):
        # call the generic folder contents browserview. Check that it makes
        # the results a contentlisting, regardless of batching
        self.folder.invokeFactory("Document", "mypage")
        contentlisting = self.folder.restrictedTraverse("@@contentlisting")(
            batch=True,
            b_size=1,
        )
        self.assertTrue(verifyObject(IContentListing, contentlisting))
        self.assertEqual(len(contentlisting), 1)

    def test_batching_folder_contents_2(self):
        # call the generic folder contents browserview. Check that it makes
        # the results a contentlisting, regardless of batching
        new_id = self.folder.invokeFactory("Document", "mypage")
        new_id2 = self.folder.invokeFactory("Document", "mypage2")
        contentlisting = self.folder.restrictedTraverse("@@contentlisting")(
            batch=True,
            b_size=1,
        )
        self.assertTrue(contentlisting[0].getId() == new_id)
        self.assertEqual(len(contentlisting), 1)
        self.assertEqual(contentlisting.actual_result_count, 2)

        contentlisting = self.folder.restrictedTraverse("@@contentlisting")(
            batch=True,
            b_size=1,
            b_start=1,
        )
        self.assertEqual(contentlisting[0].getId(), new_id2)
        self.assertEqual(len(contentlisting), 1)
        self.assertEqual(contentlisting.actual_result_count, 2)


class TestCollectionResults(unittest.TestCase):
    """Test, if the @@contentlisting view also works for Collections."""

    layer = CONTENTLISTING_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()
        self.portal = self.layer["portal"]
        self.folder = self.portal["test-folder"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Collection", "collection", title="Col")
        collection = self.portal.collection
        collection.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.selection.any",
                "v": ["Event", "Event"],
            },
        ]
        collection.reindexObject()
        self.col = collection

    def test_collection_results_is_contentlisting(self):
        # call the generic contentlisting view. Check that it makes results
        # a contentlisting, regardless of batching
        self.folder.invokeFactory("Event", "myevent")
        contentlisting = self.col.restrictedTraverse("@@contentlisting")()

        self.assertTrue(verifyObject(IContentListing, contentlisting))

    def test_filtering_collection_results_to_empty(self):
        contentlisting = self.col.restrictedTraverse("@@contentlisting")(
            portal_type="NotExistent",
        )

        self.assertEqual(len(contentlisting), 0)
        self.assertEqual(contentlisting.actual_result_count, 0)

    def test_filtering_collection_results_to_news_items(self):
        self.folder.invokeFactory("Link", "mylink")
        contentlisting = self.col.restrictedTraverse("@@contentlisting")(
            portal_type="Link",
        )

        self.assertEqual(len(contentlisting), 1)
        self.assertEqual(contentlisting.actual_result_count, 1)
        self.assertEqual(contentlisting[0].portal_type, "Link")

    def test_item_in_collection_results(self):
        self.folder.invokeFactory("Event", "myevent")
        contentlisting = self.col.restrictedTraverse("@@contentlisting")()

        self.assertEqual(len(contentlisting), 1)
        self.assertEqual(contentlisting.actual_result_count, 1)

    def test_batching_collection_results(self):
        # call the contentlisting view. Check that it makes
        # the results a contentlisting, regardless of batching
        self.folder.invokeFactory("Event", "myevent")
        contentlisting = self.col.restrictedTraverse("@@contentlisting")(
            batch=True,
            b_size=1,
        )

        # In case of Collections, the result is a plone.batching object
        self.assertTrue(IBatch.providedBy(contentlisting))
        self.assertTrue(verifyObject(IContentListingObject, contentlisting[0]))
        self.assertEqual(len(contentlisting), 1)

    def test_batching_collection_results_2(self):
        # call the contentlisting view. Check that it makes
        # the results a contentlisting, regardless of batching
        new_id = self.folder.invokeFactory("Event", "myevent")
        new_id2 = self.folder.invokeFactory("Event", "myevent2")
        contentlisting = self.col.restrictedTraverse("@@contentlisting")(
            batch=True,
            b_size=1,
        )

        self.assertTrue(contentlisting[0].getId() == new_id)
        self.assertEqual(contentlisting.items_on_page, 1)
        self.assertEqual(len(contentlisting), 2)
        self.assertEqual(contentlisting.has_next, True)

        contentlisting = self.col.restrictedTraverse("@@contentlisting")(
            batch=True,
            b_size=1,
            b_start=1,
        )

        self.assertEqual(contentlisting[0].getId(), new_id2)
        self.assertEqual(contentlisting.items_on_page, 1)
        self.assertEqual(len(contentlisting), 2)
        self.assertEqual(contentlisting.has_next, False)


def test_suite():
    import unittest

    suite = unittest.TestSuite()
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestSetup))
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(
            TestIndividualCatalogContentItems
        )
    )
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestIndividualRealContentItems)
    )
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFolderContents))
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestCollectionResults)
    )
    return suite



from plone.app.contentlisting.interfaces import IContentListing
from Products.CMFCore.utils import getToolByName
from zope.publisher.browser import BrowserView


class FolderListing(BrowserView):
    def __call__(self, batch=False, b_size=20, b_start=0, orphan=0, **kw):
        query = {}
        query.update(kw)

        query["path"] = {
            "query": "/".join(self.context.getPhysicalPath()),
            "depth": 1,
        }

        # if we don't have asked explicitly for other sorting, we'll want
        # it by position in parent
        if "sort_on" not in query:
            query["sort_on"] = "getObjPositionInParent"

        # Provide batching hints to the catalog
        if batch:
            query["b_start"] = b_start
            query["b_size"] = b_size + orphan

        catalog = getToolByName(self.context, "portal_catalog")
        results = catalog(query)
        return IContentListing(results)


class ContentListingCollection(BrowserView):
    def __call__(self, batch=False, b_size=20, b_start=0, **kw):
        if "orphan" in kw:
            # At the moment, orphan keyword is not supported by
            # plone.app.contenttypes Collection behavior, nor by
            # plone.app.querystring's querybuilder.
            del kw["orphan"]

        res = self.context.results(
            batch=batch,
            b_size=b_size,
            b_start=b_start,
            custom_query=kw,
        )
        return res



from Acquisition import aq_get
from plone.app.contentlisting.contentlisting import BaseContentListingObject
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.base.interfaces import IImageScalesAdapter
from plone.base.utils import base_hasattr
from plone.base.utils import human_readable_size
from plone.rfc822.interfaces import IPrimaryFieldInfo
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
from zope.interface import implementer


MARKER = object()


@implementer(IContentListingObject)
class RealContentListingObject(BaseContentListingObject):
    """A content object representation wrapping a real content object."""

    def __init__(self, obj):
        self._realobject = obj
        self.request = aq_get(obj, "REQUEST")

    def __repr__(self):
        return (
            "<plone.app.contentlisting.realobject."
            f"RealContentListingObject instance at {self.getPath()}>"
        )

    __str__ = __repr__

    def __getattr__(self, name):
        """We'll override getattr so that we can defer name lookups to
        the real underlying objects without knowing the names of all
        attributes.
        """
        if name.startswith("_"):
            raise AttributeError(name)
        obj = self.getObject()
        # we check that obj without acquisition has the attribute
        # but we return it with acquisition in case it is a method
        # that itself uses the acquisition
        if base_hasattr(obj, name):
            return getattr(obj, name)
        raise AttributeError(name)

    def getObject(self):
        return self._realobject

    def getDataOrigin(self):
        # The origin of the data for the object.
        # Sometimes we just need to know if we are looking at a brain or
        # the real object.
        return self.getObject()

    # a base set of elements that are needed but not defined in dublin core
    def getPath(self):
        return "/".join(self.getObject().getPhysicalPath())

    def getURL(self):
        return self.getObject().absolute_url()

    def uuid(self):
        # content objects might have UID and might not. Same thing for
        # their brain.
        uuid = IUUID(self.getObject(), None)
        if uuid is not None:
            return uuid
        return self.getPath()

    def getSize(self):
        # return size of the primary field content.
        # this works the same way as the indexer works
        obj = self.getObject()
        try:
            primary_field_info = IPrimaryFieldInfo(obj, None)
        except TypeError:
            # no primary field available, dexterity raises a TypeError
            # with the slightly misleading message 'could not adapt'.
            primary_field_info = None
        if primary_field_info is None or not primary_field_info.value:
            size = 0
        else:
            size = getattr(primary_field_info.value, "size", 0)
        return human_readable_size(size)

    def review_state(self):
        obj = self.getObject()
        wftool = getToolByName(obj, "portal_workflow")
        return wftool.getInfoFor(obj, "review_state", default=None)

    def Type(self):
        # Dublin Core element - Object type.
        obj = self.getObject()
        typestool = getToolByName(obj, "portal_types")
        ti = typestool.getTypeInfo(obj)
        if ti is not None:
            return ti.Title()
        return obj.meta_type

    # Needed: A method Type() that returns the same as is cataloged as Type.
    # Currently Type() returns different values depending on the data source being
    # a brain or a real object. Probably needed. Support for all the attributes
    # from the indexablemetadata wrappers.

    def PortalType(self):
        obj = self.getObject()
        return obj.portal_type

    def image_scales(self):
        obj = self.getObject()
        return getMultiAdapter((obj, self.request), IImageScalesAdapter)()



from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.CMFCore.utils import getToolByName


class ContentListingLayer(PloneSandboxLayer):
    defaultBases = (PLONE_APP_CONTENTTYPES_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        import plone.app.contentlisting

        self.loadZCML(package=plone.app.contentlisting)


CONTENTLISTING_FIXTURE = ContentListingLayer()


class ContentListingIntegrationLayer(PloneSandboxLayer):
    defaultBases = (CONTENTLISTING_FIXTURE,)

    def setUpPloneSite(self, portal):
        setRoles(portal, TEST_USER_ID, ["Manager"])
        wftool = getToolByName(portal, "portal_workflow")
        wftool.setDefaultChain("simple_publication_workflow")

        portal.invokeFactory("Folder", "test-folder")
        portal.invokeFactory("Document", "front-page")
        portal.invokeFactory("Folder", "news")
        wftool.doActionFor(portal.news, "publish")
        portal.news.invokeFactory("News Item", "news1")
        setRoles(portal, TEST_USER_ID, ["Member"])
        from Products.CMFCore.indexing import processQueue

        processQueue()


CONTENTLISTING_INTEGRATION_FIXTURE = ContentListingIntegrationLayer()
CONTENTLISTING_INTEGRATION_TESTING = IntegrationTesting(
    bases=(CONTENTLISTING_INTEGRATION_FIXTURE,),
    name="ContentListing:Integration",
)
CONTENTLISTING_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(CONTENTLISTING_INTEGRATION_FIXTURE,),
    name="ContentListing:Functional",
)



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "3.0.6"

long_description = (
    f"{Path('README.rst').read_text()}\n{Path('CHANGES.rst').read_text()}\n"
)

setup(
    name="plone.app.contentlisting",
    version=version,
    description="Listing of content for the Plone CMS",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: Core",
        "Framework :: Plone :: 6.0",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="content list Plone",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.app.contentlisting",
    license="GPL version 2",
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "setuptools",
        "plone.base",
        "Products.MimetypesRegistry",
        "Products.ZCatalog",
        "plone.i18n",
        "plone.registry",
        "plone.rfc822",
        "plone.uuid",
        "Zope",
    ],
    extras_require={
        "test": [
            "plone.app.contenttypes[test]",
            "plone.app.testing",
            "plone.batching",
            "plone.namedfile",
            "plone.testing",
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
